@@ -1,5 +1,5 @@
 # --- BEGIN COPYRIGHT BLOCK ---
-# Copyright (C) 2015 Red Hat, Inc.
+# Copyright (C) 2022 Red Hat, Inc.
 # All rights reserved.
 #
 # License: GPL (version 3 or any later version).
@@ -7,24 +7,25 @@
 # --- END COPYRIGHT BLOCK ---
 
 from datetime import datetime
+import os
 import copy
 import ldap
-from lib389._constants import *
-from lib389.properties import *
+from lib389._constants import DN_LDBM, DN_CHAIN, DN_PLUGIN, DEFAULT_BENAME
+from lib389.properties import BACKEND_OBJECTCLASS_VALUE, BACKEND_PROPNAME_TO_ATTRNAME, BACKEND_CHAIN_BIND_DN, \
+                              BACKEND_CHAIN_BIND_PW, BACKEND_CHAIN_URLS, BACKEND_PROPNAME_TO_ATTRNAME, BACKEND_NAME, \
+                              BACKEND_SUFFIX, BACKEND_SAMPLE_ENTRIES, TASK_WAIT
 from lib389.utils import normalizeDN, ensure_str, assert_c
 from lib389 import Entry
 
 # Need to fix this ....
 
-from lib389._mapped_object import DSLdapObjects, DSLdapObject
+from lib389._mapped_object import DSLdapObjects, DSLdapObject, CompositeDSLdapObject
 from lib389.mappingTree import MappingTrees
 from lib389.exceptions import NoSuchEntryError, InvalidArgumentError
-from lib389.replica import Replicas
+from lib389.replica import Replicas, Changelog
 from lib389.cos import (CosTemplates, CosIndirectDefinitions,
                         CosPointerDefinitions, CosClassicDefinitions)
 
-# We need to be a factor to the backend monitor
-from lib389.monitor import MonitorBackend
 from lib389.index import Index, Indexes, VLVSearches, VLVSearch
 from lib389.tasks import ImportTask, ExportTask, Tasks
 from lib389.encrypted_attributes import EncryptedAttr, EncryptedAttrs
@@ -33,7 +34,7 @@ from lib389.encrypted_attributes import EncryptedAttr, EncryptedAttrs
 # This is for sample entry creation.
 from lib389.configurations import get_sample_entries
 
-from lib389.lint import DSBLE0001, DSBLE0002, DSBLE0003, DSVIRTLE0001
+from lib389.lint import DSBLE0001, DSBLE0002, DSBLE0003, DSBLE0004, DSBLE0005, DSBLE0006, DSVIRTLE0001, DSCLLE0001
 
 
 class BackendLegacy(object):
@@ -425,13 +426,14 @@ class Backend(DSLdapObject):
 
     def _lint_virt_attrs(self):
         """Check if any virtual attribute are incorrectly indexed"""
+        bename = self.lint_uid()
         indexes = self.get_indexes()
         suffix = self.get_attr_val_utf8('nsslapd-suffix')
-
         # First check nsrole
         try:
             indexes.get('nsrole')
             report = copy.deepcopy(DSVIRTLE0001)
+            report['check'] = f'backends:{bename}:virt_attrs'
             report['detail'] = report['detail'].replace('ATTR', 'nsrole')
             report['fix'] = report['fix'].replace('ATTR', 'nsrole')
             report['fix'] = report['fix'].replace('SUFFIX', suffix)
@@ -455,6 +457,7 @@ class Backend(DSLdapObject):
                         indexes.get(attr)
                         # If we got here there is an index (bad)
                         report = copy.deepcopy(DSVIRTLE0001)
+                        report['check'] = f'backends:{bename}:virt_attrs'
                         report['detail'] = report['detail'].replace('ATTR', attr)
                         report['fix'] = report['fix'].replace('ATTR', attr)
                         report['fix'] = report['fix'].replace('SUFFIX', suffix)
@@ -471,16 +474,19 @@ class Backend(DSLdapObject):
         """Perform a search and make sure an entry is accessible
         """
         dn = self.get_attr_val_utf8('nsslapd-suffix')
+        bename = self.lint_uid()
         suffix = DSLdapObject(self._instance, dn=dn)
         try:
             suffix.get_attr_val('objectclass')
         except ldap.NO_SUCH_OBJECT:
             # backend root entry not created yet
             DSBLE0003['items'] = [dn, ]
+            DSBLE0003['check'] = f'backends:{bename}:search'
             yield DSBLE0003
         except ldap.LDAPError as e:
             # Some other error
             DSBLE0002['detail'] = DSBLE0002['detail'].replace('ERROR', str(e))
+            DSBLE0002['check'] = f'backends:{bename}:search'
             DSBLE0002['items'] = [dn, ]
             yield DSBLE0002
 
@@ -491,19 +497,116 @@ class Backend(DSLdapObject):
         * missing mapping tree entries for the backend
         * missing indices if we are local and have log access?
         """
-
         # Check for the missing mapping tree.
-        suffix = self.get_attr_val_utf8('nsslapd-suffix')
-        bename = self.get_attr_val_utf8('cn')
+        suffix = self.get_attr_val_utf8_l('nsslapd-suffix')
+        bename = self.lint_uid()
         try:
             mt = self._mts.get(suffix)
-            if mt.get_attr_val_utf8('nsslapd-backend') != bename and mt.get_attr_val_utf8('nsslapd-state') != 'backend':
+            if mt.get_attr_val_utf8_l('nsslapd-backend') != bename.lower() and mt.get_attr_val_utf8('nsslapd-state') != 'backend':
                 raise ldap.NO_SUCH_OBJECT("We have a matching suffix, but not a backend or correct database name.")
         except ldap.NO_SUCH_OBJECT:
             result = DSBLE0001
+            result['check'] = f'backends:{bename}:mappingtree'
             result['items'] = [bename, ]
             yield result
 
+
+    def _lint_backend_implementation(self):
+        backend_impl = self._instance.get_db_lib()
+        if backend_impl == 'bdb':
+            result = DSBLE0006
+            result['items'] = [self.lint_uid()]
+            yield result
+
+
+    def _lint_backend_implementation_cleanup_needed(self):
+        db_files = os.listdir(self._instance.ds_paths.db_dir)
+        if self._instance.ds_paths.db_home_dir is not None and self._instance.ds_paths.db_home_dir != self._instance.ds_paths.db_dir:
+            db_files.extend(os.listdir(self._instance.ds_paths.db_home_dir))
+        db_files = [f.lower() for f in db_files]
+        db_files = sorted(set(db_files))
+
+        mdb_files = ['data.mdb', 'info.mdb', 'lock.mdb']
+        bdb_files = ['__db.001', 'dbversion', '__db.003', 'userroot', 'log.0000000001', '__db.002']
+
+        bdb_files_present = all(file in db_files for file in bdb_files)
+        mdb_files_present = all(file in db_files for file in mdb_files)
+
+        if bdb_files_present and mdb_files_present:
+            result = DSBLE0004
+            result['items'] = [self.lint_uid()]
+            yield result
+
+    def _lint_backend_implementation_config_attributes(self):
+        backend_impl = self._instance.get_db_lib()
+        db_config = DatabaseConfig(self._instance)
+        config_attrs = db_config.get()
+
+        mdb_only_attrs = ['nsslapd-mdb-max-size', 'nsslapd-mdb-max-readers', 'nsslapd-mdb-max-dbs']
+        bdb_only_attrs = ['nsslapd-dbcachesize',
+                          'nsslapd-dbncache',
+                          'nsslapd-db-logdirectory',
+                          'nsslapd-db-circular-logging',
+                          'nsslapd-db-transaction-wait',
+                          'nsslapd-db-checkpoint-interval',
+                          'nsslapd-db-compactdb-interval',
+                          'nsslapd-db-compactdb-time',
+                          'nsslapd-db-transaction-batch-val',
+                          'nsslapd-db-transaction-batch-min-wait',
+                          'nsslapd-db-transaction-batch-max-wait',
+                          'nsslapd-db-logbuf-size',
+                          'nsslapd-db-page-size',
+                          'nsslapd-db-index-page-size',
+                          'nsslapd-db-logfile-size',
+                          'nsslapd-db-trickle-percentage',
+                          'nsslapd-db-spin-count',
+                          'nsslapd-db-verbose',
+                          'nsslapd-db-debug',
+                          'nsslapd-db-locks',
+                          'nsslapd-db-locks-monitoring-enabled',
+                          'nsslapd-db-locks-monitoring-threshold',
+                          'nsslapd-db-locks-monitoring-pause',
+                          'nsslapd-db-named-regions',
+                          'nsslapd-db-private-mem',
+                          'nsslapd-db-private-import-mem',
+                          'nsslapd-db-shm-key',
+                          'nsslapd-db-debug-checkpointing',
+                          'nsslapd-db-lockdown',
+                          'nsslapd-db-tx-max',
+                          'nsslapd-online-import-encrypt',
+                          'nsslapd-db-deadlock-policy']
+
+        if backend_impl == 'bdb':
+            invalid_attrs = [attr for attr in mdb_only_attrs if attr in config_attrs]
+        else:  # mdb
+            invalid_attrs = [attr for attr in bdb_only_attrs if attr in config_attrs]
+
+        if invalid_attrs:
+            result = DSBLE0005
+            result['items'] = invalid_attrs
+            yield result
+
+    def _lint_cl_trimming(self):
+        """Check that cl trimming is at least defined to prevent unbounded growth"""
+        bename = self.lint_uid()
+        suffix = self.get_attr_val_utf8('nsslapd-suffix')
+        replicas = Replicas(self._instance)
+        try:
+            # Check if replication is enabled
+            replicas.get(suffix)
+            # Check the changelog
+            cl = Changelog(self._instance, suffix=suffix)
+            if cl.get_attr_val_utf8('nsslapd-changelogmaxentries') is None and \
+               cl.get_attr_val_utf8('nsslapd-changelogmaxage') is None:
+                report = copy.deepcopy(DSCLLE0001)
+                report['fix'] = report['fix'].replace('YOUR_INSTANCE', self._instance.serverid)
+                report['check'] = f'backends:{bename}::cl_trimming'
+                yield report
+        except:
+            # Suffix is not replicated
+            self._log.debug(f"_lint_cl_trimming - backend ({suffix}) is not replicated")
+            pass
+
     def create_sample_entries(self, version):
         """Creates sample entries under nsslapd-suffix value
 
@@ -559,7 +662,7 @@ class Backend(DSLdapObject):
 
         return (dn, valid_props)
 
-    def create(self, dn=None, properties=None, basedn=DN_LDBM):
+    def create(self, dn=None, properties=None, basedn=DN_LDBM, create_mapping_tree=True):
         """Add a new backend entry, create mapping tree,
          and, if requested, sample entries
 
@@ -569,6 +672,8 @@ class Backend(DSLdapObject):
         :type properties: dict
         :param basedn: Base DN of the new entry
         :type basedn: str
+        :param create_mapping_tree: If a related mapping tree node should be created
+        :type create_mapping_tree: bool
 
         :returns: DSLdapObject of the created entry
         """
@@ -582,8 +687,7 @@ class Backend(DSLdapObject):
             dn = ",".join(dn_comps)
 
         if properties is not None:
-            suffix_dn = properties['nsslapd-suffix'].lower()
-            dn_comps = ldap.dn.explode_dn(suffix_dn)
+            dn_comps = ldap.dn.explode_dn(properties['nsslapd-suffix'])
             ndn = ",".join(dn_comps)
             properties['nsslapd-suffix'] = ndn
             sample_entries = properties.pop(BACKEND_SAMPLE_ENTRIES, False)
@@ -593,16 +697,19 @@ class Backend(DSLdapObject):
         super(Backend, self).create(dn, properties, basedn)
 
         # We check if the mapping tree exists in create, so do this *after*
-        properties = {
-            'cn': self._nprops_stash['nsslapd-suffix'],
-            'nsslapd-state': 'backend',
-            'nsslapd-backend': self._nprops_stash['cn'],
-        }
-        if parent_suffix:
-            # This is a subsuffix, set the parent suffix
-            properties['nsslapd-parent-suffix'] = parent_suffix
-        self._mts.create(properties=properties)
-        if sample_entries is not False:
+        if create_mapping_tree is True:
+            properties = {
+                'cn': self._nprops_stash['nsslapd-suffix'],
+                'nsslapd-state': 'backend',
+                'nsslapd-backend': self._nprops_stash['cn'],
+            }
+            if parent_suffix:
+                # This is a subsuffix, set the parent suffix
+                properties['nsslapd-parent-suffix'] = parent_suffix
+            self._mts.create(properties=properties)
+
+        # We can't create the sample entries unless a mapping tree was installed.
+        if sample_entries is not False and create_mapping_tree is True:
             self.create_sample_entries(sample_entries)
         return self
 
@@ -623,7 +730,7 @@ class Backend(DSLdapObject):
             mt = self._mts.get(selector=bename)
             # Assert the type is "backend"
             # Are these the right types....?
-            if mt.get_attr_val_utf8('nsslapd-state').lower() != 'backend':
+            if mt.get_attr_val_utf8_l('nsslapd-state') != 'backend':
                 raise ldap.UNWILLING_TO_PERFORM('Can not delete the mapping tree, not for a backend! You may need to delete this backend via cn=config .... ;_; ')
 
             # Delete replicas first
@@ -663,6 +770,8 @@ class Backend(DSLdapObject):
 
     def get_monitor(self):
         """Get a MonitorBackend(DSLdapObject) for the backend"""
+        # We need to be a factor to the backend monitor
+        from lib389.monitor import MonitorBackend
 
         monitor = MonitorBackend(instance=self._instance, dn="cn=monitor,%s" % self._dn)
         return monitor
@@ -675,20 +784,20 @@ class Backend(DSLdapObject):
 
     def get_index(self, attr_name):
         for index in self.get_indexes().list():
-            idx_name = index.get_attr_val_utf8_l('cn').lower()
+            idx_name = index.get_attr_val_utf8_l('cn')
             if idx_name == attr_name.lower():
                 return index
         return None
 
     def del_index(self, attr_name):
         for index in self.get_indexes().list():
-            idx_name = index.get_attr_val_utf8_l('cn').lower()
+            idx_name = index.get_attr_val_utf8_l('cn')
             if idx_name == attr_name.lower():
                 index.delete()
                 return
         raise ValueError("Can not delete index because it does not exist")
 
-    def add_index(self, attr_name, types, matching_rules=[], reindex=False):
+    def add_index(self, attr_name, types, matching_rules=None, reindex=False):
         """ Add an index.
 
         :param attr_name - name of the attribute to index
@@ -696,6 +805,21 @@ class Backend(DSLdapObject):
         :param matching_rules - a List of matching rules for the index
         :param reindex - If set to True then index the attribute after creating it.
         """
+
+        # Reject adding an index for a virtual attribute
+        virt_attr_list = ['nsrole']
+        for cosDefType in [CosIndirectDefinitions, CosPointerDefinitions, CosClassicDefinitions]:
+            defs = cosDefType(self._instance, self.get_suffix()).list()
+            for cosDef in defs:
+                attrs = cosDef.get_attr_val_utf8_l("cosAttribute").split()
+                for attr in attrs:
+                    if attr in ["default", "override", "operational", "operational-default", "merge-schemes"]:
+                        # We are at the end, just break out
+                        break
+                    virt_attr_list.append(attr)
+        if attr_name.lower() in virt_attr_list:
+            raise ValueError(f"You should not index a virtual attribute ({attr_name})")
+
         new_index = Index(self._instance)
         props = {'cn': attr_name,
                  'nsSystemIndex': 'False',
@@ -705,7 +829,9 @@ class Backend(DSLdapObject):
             mrs = []
             for mr in matching_rules:
                 mrs.append(mr)
-            props['nsMatchingRule'] = mrs
+            # Only add if there are actually rules present in the list.
+            if len(mrs) > 0:
+                props['nsMatchingRule'] = mrs
         new_index.create(properties=props, basedn="cn=index," + self._dn)
 
         if reindex:
@@ -719,7 +845,7 @@ class Backend(DSLdapObject):
         args = None
         if wait:
             args = {TASK_WAIT: True}
-        bename = ensure_str(self.get_attr_val_bytes('cn'))
+        bename = self.get_attr_val_utf8('cn')
         reindex_task = Tasks(self._instance)
         reindex_task.reindex(benamebase=bename, attrname=attrs, args=args)
 
@@ -784,7 +910,7 @@ class Backend(DSLdapObject):
 
         # return specific search
         for vlv in vlv_searches:
-            search_name = vlv.get_attr_val_utf8_l('cn').lower()
+            search_name = vlv.get_attr_val_utf8_l('cn')
             if search_name == vlv_name.lower():
                 return vlv
 
@@ -806,17 +932,17 @@ class Backend(DSLdapObject):
         returns: a List of subsuffix entries
         """
         subsuffixes = []
-        top_be_suffix = self.get_attr_val_utf8_l('nsslapd-suffix').lower()
+        top_be_suffix = self.get_attr_val_utf8_l('nsslapd-suffix')
         mts = self._mts.list()
         for mt in mts:
             parent_suffix = mt.get_attr_val_utf8_l('nsslapd-parent-suffix')
             if parent_suffix is None:
                 continue
-            if parent_suffix.lower() == top_be_suffix:
-                child_suffix = mt.get_attr_val_utf8_l('cn').lower()
+            if parent_suffix == top_be_suffix:
+                child_suffix = mt.get_attr_val_utf8_l('cn')
                 be_insts = Backends(self._instance).list()
                 for be in be_insts:
-                    be_suffix = ensure_str(be.get_attr_val_utf8_l('nsslapd-suffix')).lower()
+                    be_suffix = be.get_attr_val_utf8_l('nsslapd-suffix')
                     if child_suffix == be_suffix:
                         subsuffixes.append(be)
                         break
@@ -834,6 +960,37 @@ class Backend(DSLdapObject):
     def get_cos_templates(self):
         return CosTemplates(self._instance, self._dn).list()
 
+    def get_state(self):
+        suffix = self.get_attr_val_utf8('nsslapd-suffix')
+        try:
+            mt = self._mts.get(suffix)
+        except ldap.NO_SUCH_OBJECT:
+            raise ValueError("Backend missing mapping tree entry, unable to get state")
+        return mt.get_attr_val_utf8('nsslapd-state')
+
+    def set_state(self, new_state):
+        new_state = new_state.lower()
+        suffix = self.get_attr_val_utf8('nsslapd-suffix')
+        try:
+            mt = self._mts.get(suffix)
+        except ldap.NO_SUCH_OBJECT:
+            raise ValueError("Backend missing mapping tree entry, unable to set configuration")
+
+        if new_state not in ['backend', 'disabled',  'referral',  'referral on update']:
+            raise ValueError(f"Invalid backend state {new_state}, value must be one of the following: 'backend', 'disabled',  'referral',  'referral on update'")
+
+        # Can not change state of replicated backend
+        replicas = Replicas(self._instance)
+        try:
+            # Check if replication is enabled
+            replicas.get(suffix)
+            raise ValueError("Can not change the backend state of a replicated suffix")
+        except ldap.NO_SUCH_OBJECT:
+            pass
+
+        # Ok, change the state
+        mt.replace('nsslapd-state', new_state)
+
 
 class Backends(DSLdapObjects):
     """DSLdapObjects that represents DN_LDBM base DN
@@ -940,6 +1097,24 @@ class Backends(DSLdapObjects):
         task = task.create(properties=task_properties)
         return task
 
+    def delete_all_dangerous(self):
+        """
+        Delete all backends. This deletes from longest to shortest suffix
+        to ensure correct delete ordering.
+        """
+        for be in sorted(self.list(), key=lambda be: len(be.get_suffix()), reverse=True):
+            be.delete()
+
+    def get_backend(self, suffix):
+        """
+        Return the backend associated with the provided suffix.
+        """
+        suffix_l = suffix.lower()
+        for be in self.list():
+            if be.get_attr_val_utf8_l('nsslapd-suffix') == suffix_l:
+                return be
+        return None
+
 
 class DatabaseConfig(DSLdapObject):
     """Backend Database configuration
@@ -970,12 +1145,14 @@ class DatabaseConfig(DSLdapObject):
             'nsslapd-search-use-vlv-index',
             'nsslapd-exclude-from-export',
             'nsslapd-serial-lock',
-            'nsslapd-subtree-rename-switch',
             'nsslapd-pagedlookthroughlimit',
             'nsslapd-pagedidlistscanlimit',
             'nsslapd-rangelookthroughlimit',
             'nsslapd-backend-opt-level',
             'nsslapd-backend-implement',
+            'nsslapd-db-durable-transaction',
+            'nsslapd-search-bypass-filter-test',
+            'nsslapd-serial-lock',
         ]
         self._db_attrs = {
             'bdb':
@@ -983,37 +1160,42 @@ class DatabaseConfig(DSLdapObject):
                     'nsslapd-dbcachesize',
                     'nsslapd-db-logdirectory',
                     'nsslapd-db-home-directory',
-                    'nsslapd-db-durable-transaction',
                     'nsslapd-db-transaction-wait',
                     'nsslapd-db-checkpoint-interval',
                     'nsslapd-db-compactdb-interval',
+                    'nsslapd-db-compactdb-time',
                     'nsslapd-db-page-size',
                     'nsslapd-db-transaction-batch-val',
                     'nsslapd-db-transaction-batch-min-wait',
                     'nsslapd-db-transaction-batch-max-wait',
                     'nsslapd-db-logbuf-size',
                     'nsslapd-db-locks',
+                    'nsslapd-db-locks-monitoring-enabled',
+                    'nsslapd-db-locks-monitoring-threshold',
+                    'nsslapd-db-locks-monitoring-pause',
                     'nsslapd-db-private-import-mem',
                     'nsslapd-import-cache-autosize',
                     'nsslapd-cache-autosize',
                     'nsslapd-cache-autosize-split',
                     'nsslapd-import-cachesize',
-                    'nsslapd-search-bypass-filter-test',
-                    'nsslapd-serial-lock',
                     'nsslapd-db-deadlock-policy',
                 ],
-            'lmdb': []
+            'mdb': [
+                    'nsslapd-mdb-max-size',
+                    'nsslapd-mdb-max-readers',
+                    'nsslapd-mdb-max-dbs',
+                ]
         }
         self._create_objectclasses = ['top', 'extensibleObject']
         self._protected = True
-        # This could be "bdb" or "lmdb", use what we have configured in the global config
+        # This could be "bdb" or "mdb", use what we have configured in the global config
         self._db_lib = self.get_attr_val_utf8_l('nsslapd-backend-implement')
         self._dn = "cn=config,cn=ldbm database,cn=plugins,cn=config"
         self._db_dn = f"cn={self._db_lib},cn=config,cn=ldbm database,cn=plugins,cn=config"
         self._globalObj = DSLdapObject(self._instance, dn=self._dn)
         self._dbObj = DSLdapObject(self._instance, dn=self._db_dn)
         # Assert there is no overlap in different config sets
-        assert_c(len(set(self._global_attrs).intersection(set(self._db_attrs['bdb']), set(self._db_attrs['lmdb']))) == 0)
+        assert_c(len(set(self._global_attrs).intersection(set(self._db_attrs['bdb']), set(self._db_attrs['mdb']))) == 0)
 
     def get(self):
         """Get the combined config entries"""
@@ -1035,6 +1217,10 @@ class DatabaseConfig(DSLdapObject):
                 vo = vo[0]
             self._instance.log.info(f'{k}: {vo}')
 
+    def get_db_lib(self):
+        """Return the backend library, bdb, mdb, etc"""
+        return self._db_lib
+
     def set(self, value_pairs):
         for attr, val in value_pairs:
             attr = attr.lower()
@@ -1044,8 +1230,56 @@ class DatabaseConfig(DSLdapObject):
             elif attr in self._db_attrs['bdb']:
                 db_config = DSLdapObject(self._instance, dn=self._db_dn)
                 db_config.replace(attr, val)
-            elif attr in self._db_attrs['lmdb']:
-                pass
+            elif attr in self._db_attrs['mdb']:
+                db_config = DSLdapObject(self._instance, dn=self._db_dn)
+                db_config.replace(attr, val)
             else:
                 # Unknown attribute
                 raise ValueError("Can not update database configuration with unknown attribute: " + attr)
+
+
+class BackendSuffixView(CompositeDSLdapObject):
+    """ Composite view between backend and mapping tree entries
+        used by: dsconf instance backend suffix ...
+    """
+
+    def __init__(self, instance, be):
+        super(BackendSuffixView, self).__init__(instance, be.dn)
+        be_args = [
+            'nsslapd-cachememsize',
+            'nsslapd-cachesize',
+            'nsslapd-dncachememsize',
+            'nsslapd-readonly',
+            'nsslapd-require-index',
+            'nsslapd-suffix'
+        ]
+        mt_args = [
+            'orphan',
+            'nsslapd-state',
+            'nsslapd-referral',
+        ]
+        mt = be._mts.get(be.get_suffix())
+        self.add_component(be, be_args)
+        self.add_component(mt, mt_args)
+
+    def get_state(self):
+        return self.get_attr_val_utf8('nsslapd-state')
+
+    def set_state(self, new_state):
+        new_state = new_state.lower()
+        suffix = self.get_attr_val_utf8('nsslapd-suffix')
+
+        if new_state not in ['backend', 'disabled',  'referral',  'referral on update']:
+            raise ValueError(f"Invalid backend state {new_state}, value must be one of the following: 'backend', 'disabled',  'referral',  'referral on update'")
+
+        # Can not change state of replicated backend
+        replicas = Replicas(self._instance)
+        try:
+            # Check if replication is enabled
+            replicas.get(suffix)
+            raise ValueError("Can not change the backend state of a replicated suffix")
+        except ldap.NO_SUCH_OBJECT:
+            pass
+
+        # Ok, change the state
+        self.set('nsslapd-state', new_state)
